@@ -1,20 +1,36 @@
 """Utility functions to download Ascend resource definitions."""
 
 import base64
+import collections
+from re import template
 import glog
 from google.protobuf import descriptor
+from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
 from google.protobuf.message import Message as ProtoMessage
 import jinja2
 import networkx as nx
 import os
 import pathlib
-from typing import Dict, List, Optional, Tuple
+from typing import Dict, Iterable, List, Optional, Tuple
 
-from ascend.sdk.applier import _component_id_map_from_list
+from ascend.protos.ascend import ascend_pb2
+from ascend.protos.schema import schema_pb2
 from ascend.sdk.builder import dataflow_from_proto, data_service_from_proto
 from ascend.sdk.client import Client
+from ascend.sdk.common import components_ordered_by_dependency, dataflows_ordered_by_dependency
 from ascend.sdk.definitions import Component, Dataflow, DataService, Definition, ReadConnector, Transform
 
+# A bit funky of an approach to support multiple versions of protobuf that use different internal container classes
+REPEATED_CONTAINERS = [RepeatedCompositeFieldContainer]
+try:
+  from google.protobuf.pyext._message import RepeatedCompositeContainer
+  REPEATED_CONTAINERS.append(RepeatedCompositeContainer)
+except:
+  pass
+
+TEMPLATES_V1 = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates", "v1")
+TEMPLATES_V2 = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates", "v2")
+
 
 class InlineCode:
   """ Represents the result of extraction of inline code from a component - such as
@@ -47,39 +63,52 @@ class InlineCode:
       return f'''pathlib.Path(os.path.join(os.path.dirname(os.path.realpath(__file__)), {rel_path_components})).read_text(encoding="utf-8")'''
 
 
-def download_component(client: Client, data_service_id: str, dataflow_id: str, component_id: str, resource_base_path: str = "."):
+def extract_inline_code(component: Component, resource_path: str, base_path: str):
+  icl = _inline_code_list_for(component, resource_path, base_path)
+  for ic in icl:
+    pathlib.Path(os.path.dirname(ic.resource_path)).mkdir(parents=True, exist_ok=True)
+    with open(ic.resource_path, "w", encoding='utf-8') as f:
+      f.write(ic.code)
+  return {ic.attribute_path: ic.loader() for ic in icl}
+
+
+def download_component(client: Client, data_service_id: str, dataflow_id: str, component_id: str, resource_base_path: str = ".", template_dir: str = None):
   """
   Downloads a Component and writes its definition to a file named `{component_id}.py` under
   `resource_base_path`. Inline code for PySpark and SQL Transforms are written as separate
   files in the same folder with the file names derived from the id of the component to
-  which the code belongs. Raises a `ValueError` if `resource_base_path` does not exist.
+  which the code belongs. Creates `resource_base_path` if `resource_base_path` does not exist.
 
   Parameters:
   - client: SDK client
   - data_service_id: DataService id
   - dataflow_id: Dataflow id
   - component_id: Component id
-  - resource_base_path: path to which Component definition will be written
-  are written
+  - resource_base_path: path to which Component definition will be written are written
+  - template_dir: directory that contains templates for rendering. Defaults to v1 template directory.
   """
-  if not os.path.isdir(resource_base_path):
+  if os.path.exists(resource_base_path) and not os.path.isdir(resource_base_path):
     raise ValueError(f"Specified resource path ({resource_base_path}) must be a directory")
-
-  df_proto = client.get_dataflow(data_service_id=data_service_id, dataflow_id=dataflow_id).data
-  dataflow = dataflow_from_proto(client, data_service_id, df_proto)
-
-  component_path = os.path.join(resource_base_path, f"{component_id}_apply.py")
-  glog.info(f"Writing component definition to ({component_path}) and component resource files to ({resource_base_path})")
-
   pathlib.Path(resource_base_path).mkdir(parents=True, exist_ok=True)
 
-  rendered_component, inline_code_list = render_component(data_service_id, dataflow, component_id, client.hostname, resource_base_path)
-  for inline_code in inline_code_list:
-    with open(inline_code.resource_path, "w", encoding="utf-8") as f:
-      f.write(inline_code.code)
+  # find our component and write it
+  dataflow = get_dataflow(client, data_service_id, dataflow_id)
+  for component in dataflow.components:
+    if component.id == component_id:
+      component_path = os.path.join(resource_base_path, f"{component_id}_apply.py")
+      glog.info(f"Writing component definition to ({component_path}) and component resource files to ({resource_base_path})")
+      with open(component_path, "w", encoding="utf-8") as f:
+        f.write(
+            _jinja_render(template_dir,
+                          'component.jinja',
+                          data_service_id=data_service_id,
+                          dataflow_id=dataflow_id,
+                          component=component,
+                          base_path=resource_base_path,
+                          hostname=client.hostname))
+        return
 
-  with open(component_path, "w", encoding="utf-8") as f:
-    f.write(rendered_component)
+  raise KeyError(f"Component with id of {component_id} not found in dataflow {data_service_id}.{dataflow_id}")
 
 
 def _rewrite_connection_ids_to_friendly_ids(connections: dict, dataflow: Dataflow):
@@ -87,191 +116,97 @@ def _rewrite_connection_ids_to_friendly_ids(connections: dict, dataflow: Dataflo
   for c in dataflow.components:
     # for transform, there is no container.
     cid = c.container.record_connection.connection_id.value if hasattr(c, 'container') else None
-    # backword compatible change, if use new sdk against old env, we may get into a case that
+    # backward compatible change, if use new sdk against old env, we may get into a case that
     # uuid to friendly map get an empty mapping value.
     friendly_id = cids_map.get(cid) if cid else None
     if friendly_id:
       c.container.record_connection.connection_id.value = friendly_id
 
 
-def download_dataflow(client: Client, data_service_id: str, dataflow_id: str, resource_base_path: str = "."):
+def _rewrite_credential_ids_to_friendly_ids(credentials: dict, dataflow: Dataflow):
+  cids_map = {x.credential.id.value: x.credential_id for x in credentials.data}
+  for c in dataflow.components:
+    cid = c.operator.spark_function.credentials_configuration.id.value if hasattr(c, 'operator') and hasattr(c.operator, 'spark_function') else None
+    friendly_id = cids_map.get(cid) if cid else None
+    if friendly_id:
+      c.operator.spark_function.credentials_configuration.id.value = friendly_id
+
+
+def download_dataflow(client: Client, data_service_id: str, dataflow_id: str, resource_base_path: str = ".", template_dir: str = None):
   """
   Downloads a Dataflow and writes its definition to a file named `{dataflow_id}.py` under
   `resource_base_path`. Inline code for Transforms and ReadConnectors are written as separate
-  files to a sub-folder - `resource_base_path`/`{dataflow_id}`/ with the file names derived
-  from the id of the component to which the code belongs. Raises a `ValueError` if
+  files to a sub-folder - `resource_base_path`/`components`/ with the file names derived
+  from the id of the component to which the code belongs. Creates `resource_base_path` if
   `resource_base_path` does not exist.
 
   Parameters:
   - client: SDK client
   - data_service_id: DataService id
   - dataflow_id: Dataflow id
-  - resource_base_path: path to which Dataflow definition and resource files
-  are written
+  - resource_base_path: path to which Dataflow definition and resource files are written
+  - template_dir: directory that contains templates for rendering. Defaults to v1 template directory.
   """
-  if not os.path.isdir(resource_base_path):
-    raise ValueError(f"Specified resource path ({resource_base_path}) must be a directory")
+  write_dataflow(client.hostname, data_service_id, get_dataflow(client, data_service_id, dataflow_id), resource_base_path, template_dir)
 
+
+def get_dataflow(client: Client, data_service_id: str, dataflow_id: str) -> Dataflow:
   df_proto = client.get_dataflow(data_service_id=data_service_id, dataflow_id=dataflow_id).data
-
   dataflow = dataflow_from_proto(client, data_service_id, df_proto)
 
   # translate connection id to friendly id
   connections = client.list_connections(data_service_id)
-
   _rewrite_connection_ids_to_friendly_ids(connections, dataflow)
 
-  dataflow_resource_path = os.path.join(resource_base_path, f"{dataflow.id}")
+  # translate credential id to friendly id
+  credentials = client.list_data_service_credentials(data_service_id)
+  _rewrite_credential_ids_to_friendly_ids(credentials, dataflow)
+
+  return dataflow
+
+
+def write_dataflow(hostname: str, data_service_id: str, dataflow: Dataflow, resource_base_path: str = ".", template_dir: str = None):
+  if os.path.exists(resource_base_path) and not os.path.isdir(resource_base_path):
+    raise ValueError(f"Specified resource path ({resource_base_path}) must be a directory")
+
   dataflow_path = os.path.join(resource_base_path, f"{dataflow.id}.py")
-  glog.info(f"Writing dataflow definition to ({dataflow_path})" f" and dataflow resource files to ({dataflow_resource_path})")
-
-  pathlib.Path(dataflow_resource_path).mkdir(parents=True, exist_ok=True)
-
-  rendered_dataflow, inline_code_list = render_dataflow(data_service_id, dataflow, client.hostname, dataflow_resource_path, resource_base_path)
-  for inline_code in inline_code_list:
-    with open(inline_code.resource_path, "w", encoding="utf-8") as f:
-      f.write(inline_code.code)
+  glog.info(f"Writing dataflow definition to {dataflow_path}")
+  pathlib.Path(resource_base_path).mkdir(parents=True, exist_ok=True)
 
   with open(dataflow_path, "w", encoding="utf-8") as f:
-    f.write(rendered_dataflow)
+    f.write(_jinja_render(template_dir, 'dataflow.jinja', data_service_id=data_service_id, dataflow=dataflow, hostname=hostname, base_path=resource_base_path))
 
 
-def download_data_service(client: Client, data_service_id: str, resource_base_path: str = "."):
+def download_data_service(client: Client, data_service_id: str, resource_base_path: str = ".", template_dir: str = None):
   """
   Downloads a DataService and writes its definition to a file named `{data_service_id}.py`
   under `resource_base_path`. Inline code for Transforms and ReadConnectors are written as separate
-  files to sub-folders - `resource_base_path`/`{dataflow_id}`/ with the file name derived from
-  the id of the component to which the code belongs. Raises `ValueError` if `resource_base_path`
-  does not exist.
+  files to sub-folders - `resource_base_path`/`{dataflow_id}`/`components`/ with the file name derived from
+  the id of the component to which the code belongs. Creates `resource_base_path` if
+  `resource_base_path` does not exist.
 
   Parameters:
   - client: SDK client
   - data_service_id: DataService id
-  - resource_base_path: base path to which DataService definition and resource
-  files are written
+  - resource_base_path: base path to which DataService definition and resource files are written
+  - template_dir: directory that contains templates for rendering. Defaults to v1 template directory.
   """
-  if not os.path.isdir(resource_base_path):
-    raise ValueError(f"Specified resource path ({resource_base_path}) must be a directory")
-
   ds_proto = client.get_data_service(data_service_id=data_service_id).data
   data_service = data_service_from_proto(client, ds_proto)
-  connections = client.list_connections(data_service_id)
-  for dataflow in data_service.dataflows:
-    _rewrite_connection_ids_to_friendly_ids(connections, dataflow)
-  write_data_service(client.hostname, data_service, resource_base_path)
+  data_service.dataflows = [get_dataflow(client, data_service_id, df.id) for df in data_service.dataflows]
+  write_data_service(client.hostname, data_service, resource_base_path, template_dir)
 
 
-def write_data_service(hostname: str, data_service: DataService, resource_base_path: str = "."):
+def write_data_service(hostname: str, data_service: DataService, resource_base_path: str = ".", template_dir: str = None):
+  if os.path.exists(resource_base_path) and not os.path.isdir(resource_base_path):
+    raise ValueError(f"Specified resource path ({resource_base_path}) must be a directory")
+
   data_service_path = os.path.join(resource_base_path, f"{data_service.id}.py")
-  dataflow_resource_paths = [os.path.join(resource_base_path, f"{dataflow.id}") for dataflow in data_service.dataflows]
-  glog.info(f"Writing data service definition to ({data_service_path})" f" and dataflow resource files to ({','.join(dataflow_resource_paths)})")
-
-  for path in dataflow_resource_paths:
-    pathlib.Path(path).mkdir(parents=True, exist_ok=True)
-
-  rendered_data_service, inline_code_list = render_data_service(data_service, hostname, resource_base_path)
-  for inline_code in inline_code_list:
-    with open(inline_code.resource_path, "w", encoding="utf-8") as f:
-      f.write(inline_code.code)
+  glog.info(f"Writing data service definition to {data_service_path}")
+  pathlib.Path(resource_base_path).mkdir(parents=True, exist_ok=True)
 
   with open(data_service_path, "w", encoding="utf-8") as f:
-    f.write(rendered_data_service)
-
-
-class InlineCodeMap:
-  """Maps component to InlineCode list."""
-  def __init__(self):
-    self._inline_code_map: Dict[Tuple[str, str], List[InlineCode]] = {}
-
-  def add(self, dataflow_id: str, component_id: str, inline_code: InlineCode):
-    if (dataflow_id, component_id) not in self._inline_code_map:
-      self._inline_code_map[(dataflow_id, component_id)] = []
-    self._inline_code_map[(dataflow_id, component_id)].append(inline_code)
-
-  def attribute_path_to_loader(self, dataflow_id: str, component_id: str) -> Dict[Tuple[str, ...], str]:
-    if (dataflow_id, component_id) in self._inline_code_map:
-      inline_code_list = self._inline_code_map[(dataflow_id, component_id)]
-      return {ic.attribute_path: ic.loader() for ic in inline_code_list}
-    else:
-      return {}
-
-
-def render_component(data_service_id: str, dataflow: Dataflow, component_id: str, hostname: str, base_path: str) -> Tuple[str, List[InlineCode]]:
-  target_component = None
-  inline_code_list: List[InlineCode] = []
-  inline_code_map = InlineCodeMap()
-
-  for component in dataflow.components:
-    if component.id == component_id:
-      target_component = component
-      icl = _inline_code_list_for(component, base_path, base_path)
-      for ic in icl:
-        inline_code_list.append(ic)
-        inline_code_map.add(dataflow.id, component.id, ic)
-  if target_component is None:
-    raise KeyError(f"Component with id of {component_id} not found in dataflow {data_service_id}.{dataflow.id}")
-
-  env = _jinja_env()
-  tmpl = env.get_template("component.jinja")
-  rendered = tmpl.render(data_service_id=data_service_id,
-                         dataflow_id=dataflow.id,
-                         component=target_component,
-                         hostname=hostname,
-                         renderer=_render_definition,
-                         proto_mods=_proto_mods,
-                         gmod_classes=_gmod_classes,
-                         inline_code_map=inline_code_map,
-                         classname_map=_classname_map())
-  return rendered, inline_code_list
-
-
-def render_dataflow(data_service_id: str, dataflow: Dataflow, hostname: str, resource_path: str, base_path: str) -> Tuple[str, List[InlineCode]]:
-  inline_code_list: List[InlineCode] = []
-  inline_code_map = InlineCodeMap()
-
-  for component in dataflow.components:
-    icl = _inline_code_list_for(component, resource_path, base_path)
-    for ic in icl:
-      inline_code_list.append(ic)
-      inline_code_map.add(dataflow.id, component.id, ic)
-
-  env = _jinja_env()
-  tmpl = env.get_template("dataflow.jinja")
-  rendered = tmpl.render(data_service_id=data_service_id,
-                         dataflow=dataflow,
-                         hostname=hostname,
-                         renderer=_render_definition,
-                         proto_mods=_proto_mods,
-                         gmod_classes=_gmod_classes,
-                         ordered_components=_components_ordered_by_dependency,
-                         inline_code_map=inline_code_map,
-                         classname_map=_classname_map())
-  return rendered, inline_code_list
-
-
-def render_data_service(data_service: DataService, hostname: str, resource_path: str) -> str:
-  inline_code_list: List[InlineCode] = []
-  inline_code_map = InlineCodeMap()
-
-  for dataflow in data_service.dataflows:
-    dataflow_resource_path = os.path.join(resource_path, f"{dataflow.id}")
-    for component in dataflow.components:
-      icl = _inline_code_list_for(component, dataflow_resource_path, resource_path)
-      for ic in icl:
-        inline_code_list.append(ic)
-        inline_code_map.add(dataflow.id, component.id, ic)
-
-  env = _jinja_env()
-  tmpl = env.get_template("data_service.jinja")
-  rendered = tmpl.render(data_service=data_service,
-                         hostname=hostname,
-                         renderer=_render_definition,
-                         proto_mods=_proto_mods,
-                         gmod_classes=_gmod_classes,
-                         ordered_components=_components_ordered_by_dependency,
-                         inline_code_map=inline_code_map,
-                         classname_map=_classname_map())
-  return rendered, inline_code_list
+    f.write(_jinja_render(template_dir, 'data_service.jinja', data_service=data_service, hostname=hostname, base_path=resource_base_path))
 
 
 _proto_mods = [
@@ -302,7 +237,9 @@ def _classname_map() -> dict:
 def _inline_code_list_for(component: Component, resource_path_prefix: str, base_path: str) -> Optional[List[InlineCode]]:
   if isinstance(component, Transform) and component.operator:
     if component.operator.HasField("spark_function") and component.operator.spark_function.executable.code.source.HasField("inline"):
-      if component.operator.spark_function.executable.code.language.HasField("sql"):
+      if component.operator.spark_function.executable.code.language.HasField("sql") or \
+        component.operator.spark_function.executable.code.language.HasField("snowflake_sql") or \
+        component.operator.spark_function.executable.code.language.HasField("bigquery_sql"):
         resource_name = f"{component.id}.sql"
       elif component.operator.spark_function.executable.code.language.HasField("python"):
         resource_name = f"{component.id}.py"
@@ -349,10 +286,23 @@ def _inline_code_list_for(component: Component, resource_path_prefix: str, base_
     return []
 
 
-def _jinja_env():
-  module_dir = os.path.dirname(os.path.abspath(__file__))
-  template_dir = os.path.join(module_dir, "templates")
-  return jinja2.Environment(loader=jinja2.FileSystemLoader(template_dir), trim_blocks=True, lstrip_blocks=True)
+def _jinja_render(template_dir, filename, **kwargs) -> str:
+  if template_dir is None:
+    template_dir = TEMPLATES_V1
+  else:
+    template_dir = os.path.abspath(template_dir)
+  env = jinja2.Environment(extensions=['jinja2.ext.do'], loader=jinja2.FileSystemLoader(template_dir), trim_blocks=True, lstrip_blocks=True)
+  return env.get_template(filename).render(classname_map=_classname_map(),
+                                           extract_inline_code=extract_inline_code,
+                                           gmod_classes=_gmod_classes,
+                                           ordered_components=components_ordered_by_dependency,
+                                           ordered_dataflows=dataflows_ordered_by_dependency,
+                                           os=os,
+                                           proto_mods=_proto_mods,
+                                           renderer=_render_definition,
+                                           template_dir=template_dir,
+                                           write_dataflow=write_dataflow,
+                                           **kwargs)
 
 
 def _render_definition(val: Definition, attribute_path: Tuple[str, ...] = (), indent=0, classname_map: dict = {}, attribute_overrides: dict = {}):
@@ -385,10 +335,14 @@ def _render_value(val, attribute_path: Tuple[str, ...], indent=0, classname_map:
   """ Renders values in Python definition form. Support is limited to
   resource definitions, proto messages, and native types.
   """
-  if isinstance(val, list):
+  if isinstance(val, list) or type(val) in REPEATED_CONTAINERS:
     return _render_array(val, attribute_path, indent, classname_map, attribute_overrides)
-  elif isinstance(val, dict):
+  elif isinstance(val, collections.abc.MutableMapping):
     return _render_map(val, attribute_path, indent, classname_map, attribute_overrides)
+  elif isinstance(val, schema_pb2.Field):
+    return _render_proto_field(val, attribute_path, indent, classname_map, attribute_overrides)
+  elif isinstance(val, ascend_pb2.Value):
+    return _render_proto_value(val, attribute_path, indent, classname_map, attribute_overrides)
   elif isinstance(val, ProtoMessage):
     return _render_message(val, attribute_path, indent, classname_map, attribute_overrides)
   elif isinstance(val, Definition):
@@ -431,6 +385,38 @@ def _render_map(mp, attribute_path: Tuple[str, ...], indent=0, classname_map: di
                      attribute_overrides=attribute_overrides)
 
 
+def _render_proto_field(field: schema_pb2.Field, attribute_path: Tuple[str, ...], indent=0, classname_map: dict = {}, attribute_overrides: dict = {}):
+  kind = field.schema.WhichOneof('details')
+  args = ''
+  if kind == 'array':
+    args += f', {_render_value(field.schema.array.element_schema, attribute_path, indent, classname_map, attribute_overrides)}'
+  elif kind == 'dictionary':
+    # We wrap these in Field objects as adding in more wrappers (this time for Schema types)
+    # seems like overkill.
+    key_schema = _render_value(schema_pb2.Field(name=None, schema=field.schema.dictionary.key_schema), attribute_path, indent, classname_map,
+                               attribute_overrides)
+    value_schema = _render_value(schema_pb2.Field(name=None, schema=field.schema.dictionary.value_schema), attribute_path, indent, classname_map,
+                                 attribute_overrides)
+    args = f", key_type={key_schema}, value_type={value_schema}"
+  elif kind == 'map':
+    args += f', {_render_value(field.schema.map.field, attribute_path, indent, classname_map, attribute_overrides)}'
+  return f"field.{kind.capitalize()}('{field.name}'{args})"
+
+
+def _render_proto_value(val: ascend_pb2.Value, attribute_path: Tuple[str, ...], indent=0, classname_map: dict = {}, attribute_overrides: dict = {}):
+  kind = val.WhichOneof('kind')
+  if kind == 'struct_value':
+    args = _render_value(val.struct_value.fields, attribute_path, indent, classname_map, attribute_overrides)
+  elif kind == 'union_value':
+    args = f"'{val.union_value.tag}'"
+    if val.union_value.value.struct_value.fields:
+      args += f', {_render_value(val.union_value.value, attribute_path, indent, classname_map, attribute_overrides)}'
+  else:
+    args = _render_value(getattr(val, kind), attribute_path, indent, classname_map, attribute_overrides)
+  t = ''.join(w.capitalize() for w in kind.split('_')[:-1])
+  return f"value.{t}({args})"
+
+
 def _render_message(message: ProtoMessage, attribute_path: Tuple[str, ...], indent=0, classname_map: dict = {}, attribute_overrides: dict = {}):
   def render_message_field(field_descriptor, val, attribute_path, indent):
     if field_descriptor.label == descriptor.FieldDescriptor.LABEL_REPEATED:
@@ -470,24 +456,3 @@ def _append(tpl: Tuple[str, ...], val: str):
 
 def _spaces_for(indent):
   return " " * indent
-
-
-def _components_ordered_by_dependency(components: List[Component]) -> List[Component]:
-  # return sorted(components, key=lambda c: c.name)
-
-  ordered: List[Component] = []
-
-  id_to_component = _component_id_map_from_list(components)
-  g = nx.DiGraph()
-  for component in components:
-    g.add_node(component.id)
-    for dep in component.dependencies():
-      # A data feed connector will not be in the component list.
-      if dep in id_to_component:
-        g.add_edge(component.id, dep)
-
-  for component_id in reversed(list(nx.topological_sort(g))):
-    component = id_to_component[component_id]
-    ordered.append(component)
-
-  return ordered
